/*
 * Copyright (C) 2006 Voice Sistem S.R.L.
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/*!
 * \file
 * \brief Kamailio presence module :: PUBLISH support
 * \ingroup presence
 */


#ifndef PUBLISH_H
#define PUBLISH_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/parser.h>
#include <time.h>
#include "../../core/str.h"
#include "../../core/dprint.h"
#include "../../core/mem/shm_mem.h"
#include "../../core/parser/msg_parser.h"
#include "presentity.h"
#include "../alias_db/alias_db.h"

void ps_presentity_db_timer_clean(unsigned int ticks, void *param);
void ps_ptable_timer_clean(unsigned int ticks, void *param);

int w_handle_publish(struct sip_msg *msg, char *str1, char *str2);
int ki_handle_publish(sip_msg_t *msg);
int ki_handle_publish_uri(sip_msg_t *msg, str *suri);
int update_hard_presentity(
		str *pres_uri, pres_ev_t *event, str *file_uri, str *filename);

#endif
