/*
 * File:   ro_fixup.h
 * Author: root
 *
 * Created on 06 April 2011, 9:51 PM
 */

#ifndef RO_FIXUP_H
#define RO_FIXUP_H

int ro_send_ccr_fixup(void **param, int param_no);

#endif /* RO_FIXUP_H */
