/*
 * Copyright (C) 2001-2003 FhG Fokus
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
/*!
 * \file
 * \brief Kamailio core :: compatibility defs for emergency roll-back if things do not work
 *
 * \ingroup core
 * Module: \ref core
 *
 * if that happens:
 * 	- a) define PRESERVE_ZT (affects most of parser)
 * 	- b) define DONT_REMOVE_ZT (affects first line)
 * 	- c) define SCRATCH
 * 	- d) undefine REMOVE_ALL_ZT (affects via)
 */


#ifndef _COMP_DEFS_H
#define _COMP_DEFS_H


#endif
