/*
 * Note: this file originally auto-generated by mib2c using
 *        $
 */
#ifndef KAMAILIONETCONFIG_H
#define KAMAILIONETCONFIG_H

/* function declarations */
void init_kamailioNetConfig(void);
Netsnmp_Node_Handler handle_kamailioNetConfUdpTransport;
Netsnmp_Node_Handler handle_kamailioNetConfUdpDtlsTransport;
Netsnmp_Node_Handler handle_kamailioNetConfTcpTransport;
Netsnmp_Node_Handler handle_kamailioNetConfTcpTlsTransport;
Netsnmp_Node_Handler handle_kamailioNetConfSctpTransport;
Netsnmp_Node_Handler handle_kamailioNetConfSctpTlsTransport;
Netsnmp_Node_Handler handle_kamailioNetConfWsTransport;
Netsnmp_Node_Handler handle_kamailioNetConfWsTlsTransport;
Netsnmp_Node_Handler handle_kamailioNetConfHttpServer;
Netsnmp_Node_Handler handle_kamailioNetConfMsrpRelay;
Netsnmp_Node_Handler handle_kamailioNetConfStunServer;
Netsnmp_Node_Handler handle_kamailioNetConfOutbound;

#endif /* KAMAILIONETCONFIG_H */