/*
 * Copyright (C) 2009 iptelorg GmbH
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __sip_msg_clone_h
#define __sip_msg_clone_h

#include "parser/msg_parser.h"

unsigned int sip_msg_clone_len(sip_msg_t *org_msg, int clone_lumps);

struct sip_msg *sip_msg_shm_clone(
		struct sip_msg *org_msg, int *sip_msg_len, int clone_lumps);

int msg_lump_cloner(struct sip_msg *pkg_msg, struct lump **add_rm,
		struct lump **body_lumps, struct lump_rpl **reply_lump);

int sip_msg_copy(sip_msg_t *imsg, sip_msg_t *omsg, unsigned int flags);

#endif /*__sip_msg_clone_h*/

/* vi: set ts=4 sw=4 tw=79:ai:cindent: */
