/**
 * Copyright (C) 2011 Daniel-Constantin Mierla (asipto.com)
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef _COREX_LIB_H_
#define _COREX_LIB_H_

#include "../../core/mod_fix.h"

int w_corex_append_branch(sip_msg_t *msg, gparam_t *pu, gparam_t *pq);
int corex_append_branch(sip_msg_t *msg, str *uri, str *qv);
int corex_send(sip_msg_t *msg, gparam_t *pu, enum sip_protos proto);
int corex_send_data(str *puri, str *psock, str *pdata);

int corex_add_alias_subdomains(char *aliasval);

int corex_register_check_self(void);

#endif
