/*
 * cryptographic secure pseudo random generation
 *
 * Copyright (C) 2019 Henning Westerholt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/* wrapper around fortuna cryptographic pseudo random number generator
 * https://en.wikipedia.org/wiki/Fortuna_(PRNG)
 */

#ifndef _cryptorand_h
#define _cryptorand_h

#include <sys/types.h>

/* seed the generator, will also use system randomness */
void cryptorand_seed(const unsigned int seed);

/* generate a 32 bit random number */
unsigned int cryptorand(void);

/* for initialization on startup or fork, will use system randomness */
unsigned int cryptorand_init(void);

#endif
