/*
 * SNMPStats Module  - Network Statistics
 *
 * Kamailio Server Net objects addition
 * Copyright (C) 2013 Edvina AB, Sollentuna, Sweden
 * Written by Olle E. Johansson
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Kamailio is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * USA
 *
 * Note: this file originally auto-generated by mib2c
 *
 */
#ifndef KAMAILIONET_H
#define KAMAILIONET_H

/* function declarations */
void init_kamailioNet(void);
Netsnmp_Node_Handler handle_kamailioNetTcpConnEstablished;
Netsnmp_Node_Handler handle_kamailioNetTcpConnFailed;
Netsnmp_Node_Handler handle_kamailioNetTcpConnReset;
Netsnmp_Node_Handler handle_kamailioNetTcpConnSuccess;
Netsnmp_Node_Handler handle_kamailioNetTcpConnOpen;
Netsnmp_Node_Handler handle_kamailioNetTcpConnPassiveOpen;
Netsnmp_Node_Handler handle_kamailioNetTcpConnReject;
Netsnmp_Node_Handler handle_kamailioNetTcpEnabled;
Netsnmp_Node_Handler handle_kamailioNetTcpMaxConns;
Netsnmp_Node_Handler handle_kamailioNetTcpAsync;
Netsnmp_Node_Handler handle_kamailioNetTcpConnTimeout;
Netsnmp_Node_Handler handle_kamailioNetTcpSendTimeout;
Netsnmp_Node_Handler handle_kamailioNetTcpConnLifetime;
Netsnmp_Node_Handler handle_kamailioNetTcpNoConnect;
Netsnmp_Node_Handler handle_kamailioNetTcpFdCache;
Netsnmp_Node_Handler handle_kamailioNetTcpAsync;
Netsnmp_Node_Handler handle_kamailioNetTcpAsyncConnWait;
Netsnmp_Node_Handler handle_kamailioNetTcpAsyncConnWqMax;
Netsnmp_Node_Handler handle_kamailioNetTcpAsyncWqMax;
Netsnmp_Node_Handler handle_kamailioNetTcpRdBufSize;
Netsnmp_Node_Handler handle_kamailioNetTcpDeferAccept;
Netsnmp_Node_Handler handle_kamailioNetTcpDelayedAck;
Netsnmp_Node_Handler handle_kamailioNetTcpSynCnt;
Netsnmp_Node_Handler handle_kamailioNetTcpLinger;
Netsnmp_Node_Handler handle_kamailioNetTcpKeepAlive;
Netsnmp_Node_Handler handle_kamailioNetTcpKeepIdle;
Netsnmp_Node_Handler handle_kamailioNetTcpKeepIntvl;
Netsnmp_Node_Handler handle_kamailioNetTcpKeepCnt;
Netsnmp_Node_Handler handle_kamailioNetTcpCrlfPing;
Netsnmp_Node_Handler handle_kamailioNetTcpAcceptAliases;
Netsnmp_Node_Handler handle_kamailioNetTcpAcceptNoCl;
Netsnmp_Node_Handler handle_kamailioNetWsConnsActive;
Netsnmp_Node_Handler handle_kamailioNetWsConnsActiveMax;
Netsnmp_Node_Handler handle_kamailioNetWsConnsFailed;
Netsnmp_Node_Handler handle_kamailioNetWsConnsClosedLocal;
Netsnmp_Node_Handler handle_kamailioNetWsConnsClosedRemote;
Netsnmp_Node_Handler handle_kamailioNetWsFramesRx;
Netsnmp_Node_Handler handle_kamailioNetWsFramesTx;
Netsnmp_Node_Handler handle_kamailioNetWsHandshakeSuccess;
Netsnmp_Node_Handler handle_kamailioNetWsHandshakeFailed;

#endif /* KAMAILIONET_H */
