#!/bin/bash

export MYSQL_USERNAME="root"
export MYSQL_PASSWORD=""
export MYSQL_DATABASE="openser"
export EXIT_FAILURE=2
export SIPP_SCEN_DIR="../"
export MI_HOST="127.0.0.1"
export KAMAILIO_HOST="127.0.0.1"
export SUBSCRIBE_WAIT_SECONDS=8
export NGREP_LOG_FILE="ngrep.log"
export SUBSCRIBERS_FILE="usernames.txt"
