/*
 * Copyright (C) 2001-2003 FhG Fokus
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
/*!
* \file
* \brief Kamailio core :: Receive msg function
* \ingroup core
* Module: \ref core
*/


#ifndef receive_h
#define receive_h

#include "ip_addr.h"

#define KSR_EVRT_RECEIVED_MESSAGE 1
#define KSR_EVRT_RECEIVED_DATAIN 2


int receive_msg(char *buf, unsigned int len, struct receive_info *ri);
int sip_check_fline(char *buf, unsigned int len);
unsigned int inc_msg_no(void);
void ksr_msg_env_reset(void);

int ksr_route_locks_set_init(void);
void ksr_route_locks_set_destroy(void);

int ksr_evrt_received(char *buf, unsigned int *len, receive_info_t *rcv_info,
		unsigned int evtype);

#endif
