/*
 * Copyright (C) 2010 iptelorg GmbH
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/** version strings.
 * @file ver.h
 * @ingroup: core
 */

#ifndef __ver_h
#define __ver_h

extern const char full_version[];
extern const char ver_name[];
extern const char ver_version[];
extern const char ver_arch[];
extern const char ver_os[];
extern const char ver_id[];
extern const char ver_compiled_time[];
extern const char ver_compiler[];
extern const char ver_flags[];
extern const char repo_ver[];
extern const char repo_hash[];
extern const char repo_state[];


#endif /*__ver_h*/

/* vi: set ts=4 sw=4 tw=79:ai:cindent: */
