/*
 * $Id$
 *
 * Copyright (C) 2013 Crocodile RCS Ltd
 * Copyright (C) 2017 ng-voice GmbH
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Exception: permission to copy, modify, propagate, and distribute a work
 * formed by combining OpenSSL toolkit software and the code in this file,
 * such as linking with software components and libraries released under
 * OpenSSL project license.
 *
 */
#ifndef AUTHEPH_MOD_H
#define AUTHEPH_MOD_H

#include "../../core/locking.h"
#include "../../core/str.h"
#include "../../modules/auth/api.h"

struct secret
{
	str secret_key;
	struct secret *prev;
	struct secret *next;
};

typedef enum
{
	AUTHEPH_USERNAME_NON_IETF = 0,
	AUTHEPH_USERNAME_IETF = 1,
} autheph_username_format_t;
extern autheph_username_format_t autheph_username_format;

typedef enum
{
	AUTHEPH_SHA1 = 0,
	AUTHEPH_SHA256 = 1,
	AUTHEPH_SHA384 = 2,
	AUTHEPH_SHA512 = 3,
} autheph_sha_alg_t;
extern autheph_sha_alg_t autheph_sha_alg;

extern auth_api_s_t eph_auth_api;

extern gen_lock_t *autheph_secret_lock;
#define SECRET_LOCK lock_get(autheph_secret_lock)
#define SECRET_UNLOCK lock_release(autheph_secret_lock)

#endif /* AUTHEPH_MOD_H */
