/*
 * Warning: This file is auto generated from a ragel syntax (rfc1918_parser.rl),
 * do not change it!
 */

#ifndef rfc1918_parser_h
#define rfc1918_parser_h


#include <sys/types.h>


unsigned int rfc1918_parser_execute(const char *str, size_t len);


#endif
