/*
 * rpc.h
 *
 *  Created on: 27 May 2014
 *      Author: jaybeepee
 */

#ifndef RPC_H_
#define RPC_H_

#include "../../core/rpc.h"
#include "config.h"

extern rpc_export_t cdp_rpc[];

#endif /* RPC_H_ */
