/*
 * kamcmd GPL license, standard disclaimer and copyright
 */

#ifndef __license_h_
#define __license_h_

#define COPYRIGHT "Copyright 2006 iptelorg GmbH"
#define DISCLAIMER \
	"This is free software with ABSOLUTELY NO WARRANTY.\n\
For details type `warranty'."


#define LICENSE \
	"    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License , or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program. If not, write to\n\
\n\
       The Free Software Foundation, Inc.\n\
       51 Franklin Street, Fifth Floor,\n\
       Boston, MA 02110-1301, USA."

#endif
