/*
 * Copyright (C) 2009 iptelorg GmbH
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __rpc_uac_h
#define __rpc_uac_h

#include "../../core/str.h"
#include "../../core/rpc.h"


int tm_rpc_response_list_init(void);
int tm_rpc_response_list_destroy(void);

void rpc_t_uac_start(rpc_t *rpc, void *c);
void rpc_t_uac_start_hex(rpc_t *rpc, void *c);
void rpc_t_uac_wait(rpc_t *rpc, void *c);
void rpc_t_uac_wait_hex(rpc_t *rpc, void *c);
void rpc_t_uac_wait_block(rpc_t *rpc, void *c);
void rpc_t_uac_wait_block_hex(rpc_t *rpc, void *c);

void rpc_t_uac_start_noack(rpc_t *rpc, void *c);
void rpc_t_uac_start_noack_hex(rpc_t *rpc, void *c);
void rpc_t_uac_wait_noack(rpc_t *rpc, void *c);
void rpc_t_uac_wait_noack_hex(rpc_t *rpc, void *c);
void rpc_t_uac_wait_block_noack(rpc_t *rpc, void *c);
void rpc_t_uac_wait_block_noack_hex(rpc_t *rpc, void *c);

void rpc_t_uac_attrs(rpc_t *rpc, void *c);

int t_uac_send(str *method, str *ruri, str *nexthop, str *send_socket,
		str *headers, str *body);

#endif /*__rpc_uac_h*/

/* vi: set ts=4 sw=4 tw=79:ai:cindent: */
